/*
 *    This file is part of NSA ( NullSearchAccess ).
 *
 *  NSA ( NullSearchAccess ) is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  NSA ( NullSearchAccess ) is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with NSA ( NullSearchAccess ).  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdlib.h>
#include <getopt.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <stdio.h>

#include "have_config.h"
#include "prototype.h"
#include "global.h"
#include "config.h"
#include "typedef.h"

int
main (int argc, char **argv)
{
  int ret;
  struct opt_s opt;
  struct stat_s stt;

  print_version ();
  program_name = argv[0];

  if (argc == 1)
    exit_program (err_no_argv);

  debug = 1;
  p_opt = &opt;
  p_stt = &stt;
  pile = NULL;
  th_argv = NULL;

  ret = decode_getopt (argc, argv);
  clean_memory ();
  return (!ret) ? EXIT_SUCCESS : EXIT_FAILURE;
}

int
decode_getopt (int argc, char **argv)
{
  int index;
  int debug_on;
  char c;
  struct proto_s proto[] =
    {
      {"ftp", "anonymous", "anonymous", 21},
      {"ftps", "anonymous", "anonymous", 21},
      {"pop", "posmaster", "123456", 110},
      {"pops", "posmaster", "123456", 995},
      {"imap", "root", "123456", 143},
      {"imaps", "root", "123456", 993},
      {"mysql", "root", "", 3306},
      {"ssh", "root", "admin", 22},
      {NULL, NULL, NULL, 0}
    };

  index = -1;
  debug_on = 0;

  memset (p_opt->ip_cur, 0, IPv4_SIZE);
  memset (p_opt->ip_end, 0, IPv4_SIZE);
  p_opt->n_thread = DEFAULT_NUMBERS_THREAD;
  p_opt->timeout = DEFAULT_TIMEOUT_VALUE;
  p_opt->proto = NULL;
  p_opt->login = NULL;
  p_opt->passwd = NULL;
  p_opt->log = 0;
  p_opt->port = -1;
  p_opt->verbose = 0;
  p_opt->ssl = 0;
  p_opt->get_ftp_write = 0;

  p_stt->start = 0;
  p_stt->stop = 0;
  p_stt->n_ip = 0;
  p_stt->n_open = 0;
  p_stt->n_found = 0;

  do
    {
      c = getopt (argc, argv, STR_OPT_LIST);
      switch (c)
  {
    case 'h':
      print_help ();
      break;
    case 'v':
      p_opt->verbose = 1;
      break;
    case 'd':
      debug_on = 1;
      break;
    case 'f':
      p_opt->log = 1;
      break;
    case 'w':
      p_opt->get_ftp_write = 1;
      break;
    case 'l':
      p_opt->login = strdup (optarg);
      break;
    case 'a':
      p_opt->passwd = strdup (optarg);
      break;
    case 'b':
      strncpy (p_opt->ip_cur, optarg, IPv4_SIZE-1);
      break;
    case 'e':
      strncpy (p_opt->ip_end, optarg, IPv4_SIZE-1);
      break;
    case 'm':
      if (!strcmp (proto[0].name, optarg))
        index = 0;
      if (!strcmp (proto[1].name, optarg))
        index = 1;
      if (!strcmp (proto[2].name, optarg))
        index = 2;
      if (!strcmp (proto[3].name, optarg))
        index = 3;
      if (!strcmp (proto[4].name, optarg))
        index = 4;
      if (!strcmp (proto[5].name, optarg))
        index = 5;
      if (!strcmp (proto[6].name, optarg))
        index = 6;
      if (!strcmp (proto[7].name, optarg))
        index = 7;
      break;
    case 'p':
      p_opt->port = xstrtol (optarg);
      if (p_opt->port < 0)
        exit_program (err_port);
      break;
    case 't':
      p_opt->timeout = xstrtol (optarg);
      if (p_opt->timeout < 0)
        exit_program (err_strtol);
      break;
    case 'n':
      p_opt->n_thread = xstrtol (optarg);
      if (p_opt->n_thread < 0)
        exit_program (err_strtol);
    break;
  }
    }
  while (c != -1);

  if (!p_opt->ip_cur[0] || !p_opt->ip_end[0])
    exit_program (err_ip_unknow);

  if (check_ip ())
    exit_program (err_ip_invalid);

  if (index < 0 || index > 7)
    exit_program (err_protocol_unknow);

  if (p_opt->get_ftp_write && index != 0 && index != 1)
    exit_program (err_no_write_proto);

  if (!p_opt->login)
    p_opt->login = xstrdup (proto[index].login);
  if (!p_opt->passwd)
    p_opt->passwd = xstrdup (proto[index].passwd);
  if (p_opt->port == -1)
    p_opt->port = proto[index].port;
 
  if (!strcmp (proto[index].name, "ftps") || !strcmp (proto[index].name, "pops") ||
        !strcmp (proto[index].name, "imaps"))
      p_opt->ssl = 1;
  p_opt->proto = xstrdup (proto[index].name);


  /* check presence library */
  if (index == 6)
    {
#ifndef HAVE_MYSQLLIB
      exit_program (err_no_lib_mysql);
#endif /* HAVE_MYSQLLIB */
    }
  if (index == 7)
    {
#ifndef HAVE_SSHLIB
      exit_program (err_no_lib_ssh);
#endif /* HAVE_SSHLIB */
    }

  if (p_opt->ssl)
    {
#ifndef HAVE_SSLLIB
      exit_program (err_no_lib_ssl);
#endif /* HAVE_SSLLIB */
    }

  if (p_opt->verbose)
    print_verbose_argv ();
  if (!debug_on)
    debug = 0;

  thread_config ();
  return 0;
}

int
xstrtol (const char *str)
{
  char *check = NULL;
  char *cpy = NULL;
  int nb;

  cpy = xstrdup (str);
  nb = strtol (cpy, &check, 10);
  if (*check != '\0')
    {
      if (debug)
        fprintf(stderr, "[strtol()] argument <%s> invalid.\n", str);
      nb = -1;
    }

  XFREE (&cpy);
  return nb;
}

int
check_ip (void)
{
  int tab1[4] = {0};
  int tab2[4] = {0};
  int ok[2];
  int i;

  ok[0] = parse_tab_ip (p_opt->ip_cur, tab1);
  ok[1] = parse_tab_ip (p_opt->ip_end, tab2);

  if (ok[0] == -1 || ok[1] == -1)
    return -1;

  i = 0;
  while (i < 4)
    {
      if (tab1[i] < 0 || tab1[i] > 255 || tab2[i] < 0 ||
           tab2[i] > 255 || tab1[i] > tab2[i])
        return -1;
      if (tab1[i] < tab2[i])
        return 0;
      i++;
    }

  return 0;
}

int
parse_tab_ip (const char *ip, int *tab)
{
  int i;
  char *cut = NULL;
  char copy[IPv4_SIZE] = {0};

  i = 0;
  strncpy (copy, ip, IPv4_SIZE-1);
  cut = strtok (copy, ".");
  while (cut && i < 4)
    {
      tab[i] = xstrtol (cut);
      if (tab[i++] == -1)
        return -1;
      cut = strtok (NULL, ".");
    }

  return (i < 4) ? -1 : 0;
}

int
thread_config (void)
{
  int i;
  int ok;
  pthread_attr_t attr;

  i = 0;
  ok = 1;
  signal_init ();
  pthread_attr_init (&attr);
  pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
  pthread_mutex_init (&mutex, NULL);

  th_argv = xmalloc (p_opt->n_thread * sizeof(struct th_argv_s*));
  while (i < p_opt->n_thread)
    th_argv[i++] = NULL;

  scanned_current_ip = p_opt->ip_cur;
  p_stt->start = time (NULL);
  print_started ();
  while (ok)
    {
      if (start_thread (&attr) != -1)
        {
          ok = strcmp (p_opt->ip_cur, p_opt->ip_end);
          incremente_ip (p_opt->ip_cur);
          p_stt->n_ip++;
        }
      else
        sleep(1);
    }

  i = 0;
  while (i < p_opt->n_thread)
    {
      if (th_argv[i])
        sleep(1);
      else
        i++;
    }

  p_stt->stop = time (NULL);
  print_static ();
  return 0;
}

int
start_thread (pthread_attr_t *attr)
{
  int ret;
  int index;
  pthread_t thread;

  pthread_mutex_lock (&mutex);
  index = get_thread_index ();
  if (index == -1)
    {
      pthread_mutex_unlock (&mutex);
      return -1;
    }

  th_argv[index] = xmalloc (sizeof(struct th_argv_s));
  th_argv[index]->index = index;
  strncpy(th_argv[index]->ip, p_opt->ip_cur, IPv4_SIZE-1);
  pthread_mutex_unlock (&mutex);

  ret = pthread_create (&thread, attr, (void*)fun_thread, &th_argv[index]->index);
  if (ret == -1 && debug)
    perror("[pthread] ");

  return ret;
}

void *
fun_thread (void *argu)
{
  int index;
  int ret;
  char *ftppermission = NULL;
  struct socket_s *socket = NULL;

  memcpy (&index, argu, sizeof(int));

#ifdef HAVE_MYSQLLIB
  if (p_opt->proto[0] == 'm')
    ret = nsa_mysql_connect (index);
#endif /* HAVE_MYSQLLIB */

#ifdef HAVE_SSHLIB
  if (p_opt->proto[0] == 's')
    ret = nsa_ssh_connect (index);
#endif /* HAVE_SSHLIB */

#if defined (HAVE_MYSQLLIB) || defined (HAVE_SSHLIB)
  else
    {
#endif /* HAVE_MYSQLLIB || HAVE_SSHLIB */
      socket = create_socket (th_argv[index]->ip, p_opt->port);
      if (!socket)
        {
          free_one_argv_thread (index);
          return NULL;
        }

      if (connect_socket (socket) == -1)
        {
          free_socket (&socket);
          free_one_argv_thread (index);
          return NULL;
        }

      switch (p_opt->proto[0])
        {
          case 'f':
            ret = nsa_ftp_client (socket, &ftppermission);
            break;
          case 'p':
            ret = nsa_pop3_connect (socket);
            break;
          case 'i':
            ret = nsa_imap_connect (socket);
            break;
        }
      free_socket (&socket);
      if (ret)
        ret = 2;
#if defined (HAVE_MYSQLLIB) || defined (HAVE_SSHLIB)
    }
#endif /* HAVE_MYSQLLIB && HAVE_SSHLIB */

  pthread_mutex_lock (&mutex);
  if (!ret)
    {
      p_stt->n_found++;
      p_stt->n_open++;
      push_pile (th_argv[index]->ip, ftppermission);
      if (p_opt->verbose)
        {
          printf("[verbose] -> IP/port/login/pass : %s/%d/%s/%s",
            th_argv[index]->ip, p_opt->port, p_opt->login, p_opt->passwd);
          if (p_opt->proto[0] == 'f' && p_opt->get_ftp_write)
            {
              printf("\n[verbose] -> permission /user/group/other : %c%c%c/%c%c%c/%c%c%c\n\n",
                *(ftppermission+1), *(ftppermission+2), *(ftppermission+3),
                *(ftppermission+4), *(ftppermission+5), *(ftppermission+6),
                *(ftppermission+7), *(ftppermission+8), *(ftppermission+9));
            }
        }
      if (p_opt->log)
        write_log (th_argv[index]->ip, ftppermission);
    }
  if (ret == 2)
    p_stt->n_open++;

  XFREE (&th_argv[index]);
  pthread_mutex_unlock (&mutex);
  XFREE (&ftppermission);

  return NULL;
}

void
free_one_argv_thread (const int index)
{
  pthread_mutex_lock (&mutex);
  XFREE (&th_argv[index]);
  pthread_mutex_unlock (&mutex);
}

int
get_thread_index (void)
{
  int i;

  i = 0;
  while (i < p_opt->n_thread)
    {
      if (!th_argv[i])
        return i;
      i++;
    }

  return -1;
}

int
incremente_ip (char *ip)
{
  char *cut = NULL;
  int tab[4] = {0};
  int i;

  i = 0;
  cut = strtok (ip, ".");
  while (cut)
    {
      tab[i] = atoi (cut);
      cut = strtok (NULL, ".");
      i++;
    }

  tab[3]++;
  if (tab[3] > 255)
    {
      tab[3] = 0;
      tab[2]++;
    }
  if (tab[2] > 255)
    {
      tab[2] = 0;
      tab[1]++;
    }
  if (tab[1] > 255)
    {
      tab[1] = 0;
      tab[0]++;
    }
  if (tab[0] > 255)
    return -1;

  snprintf (ip, IPv4_SIZE, "%d.%d.%d.%d",
              tab[0], tab[1], tab[2], tab[3]);
  return 0;
}

void
push_pile (const char *ip, const char *perm)
{
  struct pile_s *new = NULL;
  struct pile_s *p_pile = pile;

  new = xmalloc (sizeof(struct pile_s));
  strncpy (new->ip, ip, IPv4_SIZE);
  new->next = NULL;
  new->permission = NULL;
  if (perm)
    {
      new->permission = xcalloc (sizeof(char)*11);
      strncpy (new->permission, perm, 10);
    }

  if (!pile)
    pile = new;
  else
    {
      while (p_pile->next)
        p_pile = p_pile->next;
      p_pile->next = new;
    }
}

void
free_pile (void)
{
  struct pile_s *buff = NULL;

  while (pile)
    {
      buff = pile->next;
      XFREE (&pile->permission);
      XFREE (&pile);
      pile = buff;
    }
}

void
signal_init (void)
{
  struct sigaction act;
  int i;
  int siglist[] =
    {
      SIGUSR1, SIGUSR2, SIGINT, SIGPIPE, SIGQUIT,
      SIGTERM, SIGTSTP, SIGHUP, SIGILL, SIGABRT,
      SIGFPE, SIGSEGV, SIGALRM, SIGCHLD, SIGCONT,
      SIGTTIN, SIGTTOU, 0
    };

  i = 0;
  memset (&act, 0, sizeof(struct sigaction));
  act.sa_handler = signal_handler;

  while (siglist[i])
    sigaction (siglist[i++], &act, NULL);
}

void
signal_handler (int sig)
{
  if (sig == SIGINT)
    exit_program (signal_exit);
  if (sig == SIGPIPE && debug)
    fprintf (stderr, "recv signal SIGPIPE socket closed.\n");
  else
    print_static_current ();
}

void
write_log (const char *ip, const char *perm)
{
  FILE *file = NULL;

  file = fopen(LOG_NAME, "a");
  if (!file)
    {
      if (debug)
        perror("[open()] ");
      return;
    }

  fprintf(file, "%s%s/%s/%s\n", STR_WRITE_LOG, ip, p_opt->login, p_opt->passwd);
  if (p_opt->get_ftp_write)
    fprintf(file, "%s%c%c%c/%c%c%c/%c%c%c\n",PERM_LOG, *(perm+1), *(perm+2), *(perm+3),
      *(perm+4), *(perm+5), *(perm+6), *(perm+7), *(perm+8), *(perm+9));

  fclose (file);
}
